########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Server/Server/Commands/Init.py,v 1.57 2003/06/20 00:02:01 mgushee Exp $
"""
4Suite repository init command (4ss_manager init)

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

__doc__ = """
The init command is used to initialize servers. Initialization includes
reformatting the storage of the server, and adding initial data
structures. You will LOSE ALL DATA when you execute init on it, unless
you use the update option. You can only init servers on the machine
that the init command is executed from.  You must have write access to
where the back end stores its data (example, you must be able to write
to the postgres data base if the back end is postgres). Some servers
require user input to complete the init process.  If you are
initializing one of these servers, you will be prompted for this info.
"""

import sha, getpass, os, glob, time, sys, posixpath

from Ft import SHARE_DIR
from Ft.Lib import Uri, Wrap
from Ft.Lib.Terminfo import GetColumns
from Ft.Server import FtServerBaseException, FTSERVER_NAMESPACE
from Ft.Server.Common import AclConstants,ResourceTypes
from Ft.Server.Common import Schema, ClAuthenticate, XmlLib
from Ft.Server.Common.Install import InstallUtil, Serialize
from Ft.Server.Server import Drivers
from Ft.Server.Server import SCore, FtServerServerException, Error
from Ft.Server.Server.Drivers import PathImp
from Ft.Server.Server.Lib import ConfigFile, LogUtil
from Ft.Server.Server.SCore import RepositoryImp
from Ft.Xml.XLink import XLINK_NAMESPACE


def CommandLineInit(options,args):
    """
    Init function called from the command line tool.  Parse options and args, do validation, and then
    call the common Initialization routines.
    """

    #Gather the command line information
    do_update = options.get('update', 0)
    repoOnly = options.get('repoOnly', 0)
    #noprompt = options.get('noprompt', 0)
    confirm = options.get('confirm', 0)
    db_port = options.get('dashboard-port', 8080)

    components = args.get('components')

    #Adjust the components as needed
    if not len(components) and not repoOnly:
        #Default to full
        full = 1
    else:
        full = 0

    if repoOnly:
        #if they only want the repo, then the components can only be Core
        #If they specify anything else then it is an error.
        if components:
            sys.stderr.write("You cannot specify repo-only and a list of components\n")
            return
        components = ['Core']
    else:
        #If they didn't specify repo-only then update the component list
        if full:
            components = g_allComponentNames[:]
        components.insert(0,'Core')  #Make sure it is first.
        # data must be second
        if 'data' in components:
            components.remove('data')
            components.insert(1,'data')
        else:
            sys.stderr.write("The Data component must be first in the list; other components depend on it\n")
            return

    #Read the config file
    properties = ConfigFile.Read(options.get('config-file'))
    coreId = options.get('core-id',os.environ.get('FTSS_CORE_ID','Core'))
    try:
        properties = properties[coreId]
    except KeyError:
        raise FtServerServerException(Error.CONFIG_INVALID_CORE_ID, name=coreId)
    properties['CoreId'] = coreId


    #Validate that the user can do what they think they can
    driver = Drivers.Begin(LogUtil.NullLogger(),properties)

    exists = driver.exists()
    userName = passwd = None
    if exists == 1:
        #A repository was found.  Make sure we have access
        if do_update:
            sys.stderr.write("Updating existing repository\n")
        else:
            #Make sure the user knows
            sys.stderr.write("\nA repository currently exists!\n")
        sys.stderr.write("Please authenticate with a user name and password\n")
        sys.stderr.write("with execute permission for the init command.\n")
        userName, passwd = ClAuthenticate.UserAuthenticate(prompt="Manager",
                                                           promptOnly=1)
        driver.login(userName,passwd)
        model = driver.getSystemModel()
        stmts = model.complete(None,
                               Schema.COMMAND_FULL_NAME,
                               '4ss_manager.init')
        try:
            p = PathImp.CreateInitialPath('/',driver)
            if len(stmts) == 1:
                p = p.normalize(stmts[0].subject + ";no-traverse")
            else:
                sys.stderr.write("No Init command was found in the repo; will try /ftss/commands/4ss_manager.init\n")
                p = p.normalize('/ftss/commands/4ss_manager.init;no-traverse')
            driver.verifyAcl(p,AclConstants.EXECUTE_ACCESS,0)
        except FtServerServerException, e:
            if e.code == Error.PERMISSION_DENIED:
                raise e
            pass
        driver.rollback()
        del driver
    elif do_update:
        #You cannot update a repo that does not exist.
        sys.stderr.write("You cannot execute an update when a repository does not exist\n")
        return


    #Determine if we need to destroy the repository.  We destroy the repo is
    #a) it partially exists (exists == 0)
    #b) the user didn't specify update and exists == 1 (it is there)
    destroyRepo = (exists == 0) or (exists == 1 and not do_update)

    #Hand processing off to the common init routines
    DoInit(properties,
           userName,
           passwd,
           components,
           confirm=confirm,
           destroyRepo=destroyRepo,
           do_update=do_update)

    sys.stdout.flush()
    sys.stderr.write('\n')
    return

def DoInit(properties, userName, password, components, confirm=0,
           destroyRepo=0, users=None, quiet=0, do_update=0):
    """
    Do the actual initialization.
    Iterate over the components and do each one.
    """

    products = BuildProductList(components, destroyRepo, users, do_update=do_update)
    for (product, base) in products:
        repo = product.getRepo(userName, password, properties)
        commit = 0
        try:
            product.install(repo, base, quiet, confirm, do_update=do_update)
            commit = 1
        finally:
            if commit:
                product.commit(repo)
            else:
                product.rollback(repo)


def BuildProductList(components, destroyRepo, users, do_update=0):

    res = []
    for component in components:
        if component.lower() == 'core':
            res.append(( CoreProduct(destroyRepo,users),'/'))
        elif component.lower() == 'data':
            pl,base = BuildDataProductList()
            res.append((OtherProduct(pl,users),base))
        elif component.lower() == 'servers':
            pl,base = BuildServerList()
            res.append((OtherProduct(pl,users),base))
        elif component.lower() == 'commands':
            pl,base = BuildCommandList()
            res.append((OtherProduct(pl,users),base))
        elif component.lower() == 'docs':
            pl,base = BuildDocList(do_update=do_update)
            res.append((OtherProduct(pl,users),base))
        elif component.lower() == 'dashboard':
            path = os.path.join(SHARE_DIR, '..', 'Server', 'Server',
                                'Dashboard', 'setup.xml')
            f = open(path)
            pl = InstallUtil.Deserialize(f, Uri.OsPathToUri(path))
            f.close()
            res.append((OtherProduct(pl,users),'/ftss'))
        elif component.lower() == 'demos':
            path = os.path.join(SHARE_DIR, 'Demos', 'setup.xml')
            f = open(path)
            pl = InstallUtil.Deserialize(f, Uri.OsPathToUri(path))
            f.close()
            res.append((OtherProduct(pl,users),'/ftss'))
        else:
            sys.stderr.write('Skipping unknown component %s\n' % component)
    return res


def BuildDataProductList():
    l = []

    #Add some containers
    l.append(InstallUtil.Container('docdefs',[],None,None,None))
    l.append(InstallUtil.Container('data',[],None,None,None))
    l.append(InstallUtil.Container('data/icons',[],None,None,None))
    l.append(InstallUtil.Container('data/presmaterial',[],None,None,None))

    #stBase = posixpath.join(urllib.pathname2url(SHARE_DIR),'Data','Stylesheets')
    stBase = posixpath.join(Uri.OsPathToUri(SHARE_DIR),'Data','Stylesheets')

    #Add the docdefs
    dd = InstallUtil.XPathDocumentDefinition('docdefs/docbook1',[],None,None,None)
    dd.setPath(posixpath.join(stBase,'docbook1.docdef'))
    l.append(dd)
    dd = InstallUtil.XPathDocumentDefinition('docdefs/wsdl',[],None,None,None)
    dd.setPath(posixpath.join(stBase,'wsdl.docdef'))
    l.append(dd)
    dd = InstallUtil.XPathDocumentDefinition('docdefs/dublin_core',[],None,None,None)
    dd.setPath(posixpath.join(stBase,'dublin_core.docdef'))
    l.append(dd)

    dd = InstallUtil.XPathDocumentDefinition('docdefs/generated-docs',[],None,None,None)
    dd.setPath(posixpath.join(stBase,'generated-docs.docdef'))
    l.append(dd)

    #Get stuff from the stylesheet package
    stBase = os.path.join(SHARE_DIR, 'Data', 'Stylesheets')
    icBase = os.path.join(SHARE_DIR, 'Data', 'Icons')
    pmBase = os.path.join(SHARE_DIR, 'Data', 'Stylesheets', 'presmaterial')

    for css in ['presmaterial/presmain.css',
                'extensions.css',
                'sdocbook_html.css',
                'modules.css',
                'commandline.css',
                ]:
        r = InstallUtil.RawFile('data/' + css, [], None, 'text/css')
        r.setPath(posixpath.join(Uri.OsPathToUri(stBase), css))
        l.append(r)


    for (ext, imt_suff) in [('gif', 'gif'), ('png', 'png'), ('jpg', 'jpeg')]:
        files = glob.glob(os.path.join(stBase,'*.'+ext))
        for f in files:
            name = os.path.basename(f)
            r = InstallUtil.RawFile('data/' + name,
                                    [],
                                    None,
                                    'image/'+imt_suff)
            r.setPath(Uri.OsPathToUri(f))
            l.append(r)
        files = glob.glob(os.path.join(icBase,'*.'+ext))
        for f in files:
            name = os.path.basename(f)
            r = InstallUtil.RawFile('data/icons/' + name,
                                    [],
                                    None,
                                    'image/'+imt_suff)
            r.setPath(Uri.OsPathToUri(f))
            l.append(r)
        files = glob.glob(os.path.join(pmBase,'*.' + ext))
        for f in files:
            name = os.path.basename(f)
            r = InstallUtil.RawFile('data/presmaterial/' + name, [], None, 'image/'+ext)
            r.setPath(Uri.OsPathToUri(f))
            l.append(r)

    xmls = glob.glob(os.path.join(stBase,'*.xml')) + [os.path.join(stBase,'null')]
    for x in xmls:
        name = os.path.basename(x)
        d = InstallUtil.XmlDocument('data/'+name,[],None,None,None)
        d.setPath(Uri.OsPathToUri(x))
        l.append(d)

    stys = ['docbook_html1.xslt',
            'sdocbook_html.xslt',
            'docbook_text1.xslt',
            'identity.xslt',
            'null.xslt',
            'justify.xslt',
            'container.xslt',
            'cl-container.xslt',
            'commandline_html.xslt',
            'extensions_html.xslt',
            'modules_html.xslt',
            #'schema.xslt',
            'presentation1.xslt',
            'presentation-fo.xslt',
            'pretty.xslt',
            'xbel2rdf.xslt',
            'decorated-xml.xslt',
            'useragent.xslt',
            #'api-class.xslt',
            #'api-commandline.xslt',
            #'api-exception.xslt',
            #'api-structure.xslt',
            #'api-typedef.xslt',
            #'api.xslt',
            ]
    for s in stys:
        path = os.path.join(stBase,s)
        d = InstallUtil.XsltDocument('data/'+s,[],None,None,None)
        d.setPath(Uri.OsPathToUri(path))
        l.append(d)

    pl = InstallUtil.Product(l,name='Data')
    return pl,'/ftss'

def BuildServerList():
    l = []

    l.append(InstallUtil.Container('servers',[],None,None,None))

    serverbase = os.path.join(SHARE_DIR,'Data','Servers')

    s = InstallUtil.Server('servers/FtRpc.server',[],None,None,None)
    s.setPath(posixpath.join(Uri.OsPathToUri(serverbase),'FtRpc.server'))
    l.append(s)
    s = InstallUtil.Server('servers/Ftp.server',[],None,None,None)
    s.setPath(posixpath.join(Uri.OsPathToUri(serverbase),'Ftp.server'))
    l.append(s)
    s = InstallUtil.Server('servers/Soap.server',[],None,None,None)
    s.setPath(posixpath.join(Uri.OsPathToUri(serverbase),'Soap.server'))
    l.append(s)

    pl = InstallUtil.Product(l, name='Servers')
    return pl, '/ftss'

def BuildDocList(do_update=0):
    l = []

    l.append(InstallUtil.Container('docs', [], None, None, None))
    mapping = Serialize.g_defaultFileMap.copy()
    mapping['.doc'] = ('/ftss/docdefs/docbook1','text/xml', InstallUtil.XmlDocument)
    mapping['.xml'] = ('/ftss/docdefs/generated-docs','text/xml', InstallUtil.XmlDocument)

    termwidth = GetColumns(sys.stderr)
    docs_dir = os.path.join(SHARE_DIR,'Docs','xml')
    if os.path.exists(docs_dir):
        l.extend(Serialize.MirrorDir('/ftss/docs', docs_dir, mapping, 0))
        if not do_update and len(l) > 24:
            msg = "Note: %d docs will be installed." % len(l) + \
                  " This may be a time-consuming process. To avoid" + \
                  " this, abort now and either remove the " + docs_dir + \
                  " directory or run the init command without" + \
                  " specifying Docs installation (see --help)."
            sys.stderr.write(Wrap(msg, termwidth) + '\n')
    else:
        msg = "Note: There are no docs to " + \
              ('install','update')[do_update] + \
              ". If docs are needed, first reinstall 4Suite with" + \
              " 'setup.py install --with-docs'. Also note that" + \
              " generating the docs and installing them in the" + \
              " repository are time-consuming processes."
        sys.stderr.write(Wrap(msg, termwidth) + '\n')
    pl = InstallUtil.Product(l, name='Docs')

    return pl,'/ftss'


def BuildCommandList():
    l = []
    l.append(InstallUtil.Container('commands', [], None, None, None))
    from Ft.Server.Server.Commands import ManagerCommandLineApp
    app = ManagerCommandLineApp()

    l.extend(BuildCommandDocs(
        app, '', [(AclConstants.EXECUTE_ACCESS,AclConstants.SUPER_USER_GROUP_NAME,1)]
        ))

    from Ft.Server.Client.Commands import GeneralCommandLineApp
    app = GeneralCommandLineApp()

    l.extend(BuildCommandDocs(
        app, '', [(AclConstants.EXECUTE_ACCESS,AclConstants.WORLD_GROUP_NAME,1)]
        ))

    pl = InstallUtil.Product(l, name='Commands')
    return pl, '/ftss'


class CoreProduct(InstallUtil.Product):

    #There are two different flows through this code.
    #One, the user is doing a fresh install.  Note, this could be
    #because they want to remove an existing repo

    #Secondly, the user is doing an update.

    #The main difference is how we deal with users.  If, the user is not doing an update,
    #Then we need to add the users, and the superuser group to the installation.
    #If the user is doing an update then we don't touch these.

    coreResources = [InstallUtil.Container('ftss',[],AclConstants.SUPER_USER_GROUP_NAME,None,None),
                     InstallUtil.Container('ftss/users',[],AclConstants.SUPER_USER_GROUP_NAME,None,None),
                     InstallUtil.Container('ftss/groups',[],AclConstants.SUPER_USER_GROUP_NAME,None,None),
                     ]

    users = []
    useIndicator = 0
    def __init__(self,destroyRepo,users):
        self.destroyRepo = destroyRepo
        if users:
            self.users = users
        InstallUtil.Product.__init__(self,
                                     resourceList = self.coreResources[:],
                                     name='Repository',
                                     )

    def install(self, driver, baseDir, quiet, confirm, do_update=0):
        """
        Perform a few extra installation features.
        """
        #First, see if we need to get ride of an existing repo.
        if self.destroyRepo:
            if confirm:
                sys.stderr.write("Initialization will erase ALL DATA the 4Suite repository.\n")
                answer = raw_input('Are you sure (yes/no)? ')
                if answer.lower() not in ['yes','y']:
                    sys.stderr.write("Aborting Initialization\n")
                    return 0

            #See yah!!!
            if not quiet:
                sys.stderr.write("Removing Current Repository\n")
            driver.destroy()


        #now, if the user is not doing an update, then add the users to the resource list
        #We know the user is doing an update because we either explictly destroyed the
        #repo, or it never existed.
        if self.destroyRepo or driver.exists() == -1:
            #We are doing an update.  See if we need to ask for the user list
            if not self.users:
            #Gather  a list of users
                while 1:
                    user = raw_input('Add super user name (or just "enter" when done): ')
                    if not user:
                        if len(self.users): break
                        sys.stderr.write("You must add at least one super user!\n")
                        continue
                    same = 0
                    while not same:
                        password = getpass.getpass()
                        same = (password == getpass.getpass('Reenter password:'))
                    CoreProduct.users.append((user,password))

            #Now, add all of the users to the resource list
            for u in self.users:
                self.resourceList.append(InstallUtil.User(
                    u[0], [], None, None, None, 'ftss/users', u[1]))

            #Lastly add the super-user group
            members = map(lambda x:x[0],self.users)
            self.resourceList.append(InstallUtil.Group(
                AclConstants.SUPER_USER_GROUP_NAME, [], None, None, None,
                'ftss/groups', members))

            #Now, since we know it is not an update (inside this if statement)
            #create the new repo
            if not quiet:
                sys.stderr.write("Initializing the Repo\n")

            driver.initialize()  #Create the root

            #Update the log file
            init_time = time.asctime(time.localtime(time.time()))
            open(self.errorLogFileName,'a').write("Initialized Repo: %s\n" % init_time)
        else:
            #If the user just wants to update, then log into the driver to start a TX
            driver.login(self.userName, self.password)

        #Now, at this point in the install, we have a initialized driver so go into the normal installation
        #and add all of the system containers
        #Create a new Repo instance so the rest of install will work

        path = PathImp.CreateInitialPath('/', driver)
        repo = RepositoryImp.RepositoryImp(path, driver, None)

        return InstallUtil.Product.install(self, repo, '/', quiet, do_update=do_update)

    def getRepo(self, userName, password, properties):
        """
        For the core product, our repo is the driver
        """
        #One extra bit of trickery.  At this time we know the properties so add the errorlog to our resourceList
        self.errorLogFileName = os.path.join(
            SHARE_DIR, 'Data', 'Logs', '%s-error.log' % properties['CoreId'])

        self.coreResources = self.coreResources[:]
        self.coreResources.append(InstallUtil.UriReferenceFile(
            'ftss/error.log', [], AclConstants.SUPER_USER_GROUP_NAME,
            'text/plain', Uri.OsPathToUri(self.errorLogFileName)))

        #Store the user name and password incase we need them
        self.userName = userName
        self.password = password

        return Drivers.Begin(LogUtil.NullLogger(), properties)

    def commit(self, driver):
        driver.commit()

    def rollback(self, driver):
        pass

    def setIndicator(self):
        pass

class OtherProduct:

    def __init__(self, baseProduct, users):
        self.users = users
        self.baseProduct = baseProduct
        return

    def getRepo(self, userName, password, properties):
        if userName is None:
            if self.users is None:
                users = CoreProduct.users
                userName = CoreProduct.users[0][0]
                password = sha.new(CoreProduct.users[0][1]).hexdigest()
            else:
                userName,password = self.users[0]
        return SCore.GetRepository(userName, password, LogUtil.NullLogger(),
                                   properties)

    def install(self, repo, baseDir, quiet, confirm, do_update=0):
        if confirm:
            answer = raw_input('%s %s (yes/no)? '% (
                        ('Install', 'Update')[do_update],
                         self.baseProduct.name
                        ))
            if answer.lower() not in ['yes', 'y']:
                print "Skipping %s" % self.baseProduct.name
                return 0
        return self.baseProduct.install(repo, baseDir, quiet, do_update=do_update)

    def commit(self, repo):
        repo.txCommit()

    def rollback(self, repo):
        repo.txRollback()


class ExtendedCommand(InstallUtil.Command):

    def __init__(self, path, acl, owner, imt, docDef, command, cmdPath):
        self.command = command
        self.cmdPath = cmdPath
        InstallUtil.Command.__init__(self, path, acl, owner, imt, docDef)
        return

    def _getContent(self):
        doc = """<ftss:Command xmlns:ftss="%s" xmlns:dc="%s" xmlns:xlink="%s" name='%s' full-name='%s'>""" % (
            FTSERVER_NAMESPACE,
            Schema.DC,
            XLINK_NAMESPACE,
            self.command.name,
            posixpath.basename(self.path)
            )

        if self.command.description:
            doc = doc + '<dc:Description>%s</dc:Description>\n' % self.command.description
        if self.command.example:
            doc = doc + '<ftss:Example>%s</ftss:Example>\n' % self.command.example
        if self.command.verbose_description:
            doc = doc + '<ftss:VerboseDescription>%s</ftss:VerboseDescription>\n' % self.command.verbose_description

        doc = doc + '<ftss:Options>'

        for option in self.command.options:
            doc = doc + self.buildOptionDoc(option, self.command)

        doc = doc + '</ftss:Options>\n'

        doc = doc + '<ftss:Arguments>\n'
        for a in self.command.arguments:
            doc = doc + "<ftss:Argument name='%s' requirements = '%d'>" % (a.name,a.requirements)
            if a.description:
                doc = doc + '<dc:Description>%s</dc:Description>\n' % a.description
            doc = doc + "</ftss:Argument>"
        doc = doc + '</ftss:Arguments>\n'


        doc = doc + '<ftss:SubCommands>\n'
        for name,cmd in self.command.subCommands.items():
            doc = doc + '<ftss:CommandReference xlink:type="simple" xlink:href="%s.%s" xlink:actuate="onLoad" xlink:show="embed"/>' % (self.cmdPath,name)
        doc = doc + '</ftss:SubCommands>\n'

        doc = doc + """</ftss:Command>"""

        return XmlLib.MakeString(doc)


    def buildOptionDoc(self,option, command):
        from Ft.Lib.CommandLine import Options

        if isinstance(option,Options.ExclusiveOptions):
            doc = '<ftss:ExclusiveOption>\n'
            for o in option.choices:
                doc = doc + self.buildOptionDoc(o, command)
            doc = doc + '</ftss:ExclusiveOption>\n'

        elif isinstance(option,Options.TypedOption):
            doc = '<ftss:TypedOption '

            if option.shortName:
                doc = doc + 'short-name="%s" ' % option.shortName

            doc = doc + 'long-name="%s">\n' % option.longName
            if option.description:
                doc = doc + '<dc:Description>%s</dc:Description>\n' % option.description

            for value,desc in option.allowed:
                doc = doc + '<ftss:Allowed name="%s">' % value
                doc = doc + '<dc:Description>%s</dc:Description>\n' % desc
                doc = doc + '</ftss:Allowed>'

            if option.subOptions:
                doc = doc + '<ftss:SubOptions>\n'
                for so in command.subOptions:
                    doc = doc + self.buildOptionDoc(so, command)
                doc = doc + '</ftss:SubOptions>\n'
            doc = doc + '</ftss:TypedOption>\n'
        else:
            #Normal option
            doc = '<ftss:Option '
            if option.shortName:
                doc = doc + 'short-name="%s" ' % option.shortName

            if option.takesArg:
                doc = doc + 'arg-name="%s" ' % option.argName

            doc = doc + 'long-name="%s">\n' % option.longName
            if option.description:
                doc = doc + '<dc:Description>%s</dc:Description>\n' % option.description

            if option.subOptions:
                doc = doc + '<ftss:SubOptions>\n'
                for so in command.subOptions:
                    doc = doc + self.buildOptionDoc(so, command)
                doc = doc + '</ftss:SubOptions>\n'
            doc = doc + '</ftss:Option>\n'

        return unicode(doc)


def BuildCommandDocs(command, baseName, acl):

    if not command._fileName:
        raise "Command %s did not register a file name" % command.name

    if baseName:
        path = baseName + '.' + command.name
    else:
        path = command.name

    cmds = [ExtendedCommand('commands/'+path, acl, None, None,
                            '/ftss/docdefs/dublin_core', command, path)]
    cmds[0].setPath(Uri.OsPathToUri(command._fileName))

    for name,cmd in command.subCommands.items():
        cmds.extend(BuildCommandDocs(cmd, path, acl))

    return cmds

g_allComponents = {'Demos':"The default 4Suite Repository Demonstration applications and the server to access them",
                   'Docs':"All 4Suite documentation",
                   'Servers':'The default FtRpc and Ftp Server',
                   'Commands':'4ss and 4ss_manager commands and documentation.  These are needed to be able to use the 4ss and 4ss commandline tools',
                   'Dashboard':'The web-based control panel for 4Suite',
                   'Data':'Commonly used 4Suite data and icons (required; must be first component installed)'
                   }

g_allComponentNames = map(lambda x:x.lower(), g_allComponents.keys())

#These actually need to be sorted a bit.  Data needs to be first
g_allComponentNames.remove('data')
g_allComponentNames.insert(0, 'data')

def VerifyComponent(comp):
    if str(comp).lower() in g_allComponentNames:
        return str(comp).lower()
    raise ValueError("Component argument must be one of: %s." % ', '.join(g_allComponentNames))

g_componentHelp = """Specify which components to add/update in the repository. Allowed values are:\n\n"""

for name,value in g_allComponents.items():
    g_componentHelp += "'%s' - %s;\n" % (name.capitalize(), value)
g_componentHelp = g_componentHelp[:-2] + '.'


def Register():
    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('init',
                          'Initialize FtServer',
                          'Servers Docs Dashboard',
                          __doc__,
                          function=CommandLineInit,
                          options=Options.Options([Options.Option('c',
                                                                  'confirm',
                                                                  'Confirm before performing each sub-task.'),
                                                   Options.Option('r',
                                                                  'update',
                                                                  "Don't destroy the current repository, just update the data."),
                                                   Options.Option('o',
                                                                  'repo-only',
                                                                  'Only create the bare repository (destroying any current one), and don\'t install any components. Only useful for development.'),
                                           ]),
                          arguments = [Arguments.ZeroOrMoreArgument('components',
                                                                    g_componentHelp,
                                                                    VerifyComponent),
                                       ],

                          fileName = __file__
                          )
    return cmd




