"""
EXSLT 2.0 - Common (http://www.exslt.org/exsl/index.html)
WWW: http://4suite.org/XSLT        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, urllib
from xml.dom import Node
from Ft.Lib import boolean
from Ft.Xml.Xslt import XsltElement, OutputParameters
from Ft.Xml.Xslt import ContentInfo, AttributeInfo
from Ft.Xml.Xslt import XsltRuntimeException, Error
from Ft.Xml.Xslt.Exslt.MessageSource import Error as ExsltError

EXSL_COMMON_NS = "http://exslt.org/common"

def NodeSet(context, rtf):
    """
    The purpose of the exsl:node-set function is to convert a result tree
    fragment into a node set. If the argument is a node set already, it is
    simply returned as is. It is an error if the argument to exsl:node-set
    is not a node set or a result tree fragment.
    """
    if type(rtf) == type([]):
        return rtf
    elif getattr(rtf, 'nodeType', None) == Node.DOCUMENT_NODE:
        return [rtf]
    else:
        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
                                   context.currentInstruction)


def ObjectType(context, obj):
    """
    The exsl:object-type function returns a string giving the type of the
    object passed as the argument. The possible object types are: 'string',
    'number', 'boolean', 'node-set', 'RTF' or 'external'.
    """
    if type(obj) is type([]):
        return u'node-set'
    elif type(obj) is type(u''):
        return u'string'
    elif type(obj) in [type(1), type(2.3), type(4L)]:
        return u'number'
    elif type(obj) is type(boolean.true):
        return u'boolean'
    elif getattr(obj, 'nodeType', None) == Node.DOCUMENT_NODE:
        return u'RTF'
    else:
        return u'external'


class DocumentElement(XsltElement):
    """
    For full specification, see:
    http://www.exslt.org/exsl/elements/document/index.html
    """

    content = ContentInfo.Template
    legalAttrs = {
        'href' : AttributeInfo.UriReferenceAvt(required=1),
        'method' : AttributeInfo.QNameAvt(),
        'version' : AttributeInfo.NMTokenAvt(),
        'encoding' : AttributeInfo.StringAvt(),
        'omit-xml-declaration' : AttributeInfo.YesNoAvt(),
        'standalone' : AttributeInfo.YesNoAvt(),
        'doctype-public' : AttributeInfo.StringAvt(),
        'doctype-system' : AttributeInfo.StringAvt(),
        'cdata-section-elements' : AttributeInfo.QNamesAvt(),
        'indent' : AttributeInfo.YesNoAvt(),
        'media-type' : AttributeInfo.StringAvt(),
        }

    def __init__(self, root, namespaceUri, localName, baseUri):
        XsltElement.__init__(self, root, namespaceUri, localName, baseUri)
        self._output_parameters = OutputParameters.OutputParameters()
        return

    def instantiate(self, context, processor):
        context.processorNss = self.namespaces

        # this uses attributes directly from self
        self._output_parameters.avtParse(self, context)
        href = self._href.evaluate(context)

        #Use the main input source for relative URI rather than the main
        #"output source"
        
        uri = processor._docInputSource._normalize(href)
        scheme, rest = urllib.splittype(uri)
        #Only file URLs are supported
        if scheme != "file":
            raise XsltRuntimeException(ExsltError.UNSUPPORTED_DOCUMENT_URI_SCHEME,
                                       context.currentInstruction, scheme)
        path = urllib.url2pathname(urllib.splithost(rest)[1])
        
        try:
            stream = open(path, 'w')
        except IOError:
            dir = os.path.split(path)[0]
            #Should we also check base path writability with os.W_OK?
            if not os.access(dir, os.F_OK):
                os.makedirs(dir)
                stream = open(path, 'w')
            else:
                raise
        
        processor.addHandler(self._output_parameters, stream)
        try:
            for child in self.children:
                context = child.instantiate(context, processor)[0]
        finally:
            processor.removeHandler()
            stream.close()

        return (context,)


ExtNamespaces = {
    EXSL_COMMON_NS : 'exsl',
    }

ExtFunctions = {
    (EXSL_COMMON_NS, 'node-set'): NodeSet,
    (EXSL_COMMON_NS, 'object-type'): ObjectType,
    }

ExtElements = {
    (EXSL_COMMON_NS, 'document'): DocumentElement,
    }
